/*  $NetBSD: getopt.c,v 1.26 2003/08/07 16:43:40 agc Exp $  */

/*
 * Copyright (c) 1987, 1993, 1994
 *  The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by the University of
 *  California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <string.h>
#include "getopt.h"

#define CONST_CAST(type, p) ((type)(uintptr_t)(p))

#define BADCH   (int)'?'
#define BADARG  (int)':'
#define EMSG    ""

#ifdef WIN32
    #include <stddef.h>
    #define PATH_DELIM_CHAR    '\\'
#else
    #define PATH_DELIM_CHAR    '/'
#endif

static char* _getprogname(char* argv0)
{
    char* s;
    s = strrchr(argv0, PATH_DELIM_CHAR);
    if (s != NULL) {
        return s+1;
    }
    return argv0;
}

void
getopt_init(getopt_data_t* optdat)
{
    optdat->opterr = 1;
    optdat->optind = 1;
    optdat->optopt = 0;
    optdat->optreset = 0;
    optdat->optarg = NULL;
    optdat->place = EMSG;
}

/*
 * getopt --
 *  Parse argc/argv argument vector.
 */
int
getopt(
      getopt_data_t* optdat,
      int nargc,
      char * const nargv[],
      const char *ostr)
{
    char *oli;              /* option letter list index */

    if (optdat->optreset || *optdat->place == 0) {      /* update scanning pointer */
        optdat->optreset = 0;
        optdat->place = nargv[optdat->optind];
        if (optdat->optind >= nargc || *optdat->place++ != '-') {
            /* Argument is absent or is not an option */
            optdat->place = EMSG;
            return(-1);
        }
        optdat->optopt = *optdat->place++;
        if (optdat->optopt == '-' && *optdat->place == 0) {
            /* "--" => end of options */
            ++optdat->optind;
            optdat->place = EMSG;
            return(-1);
        }
        if (optdat->optopt == 0) {
            /* Solitary '-', treat as a '-' option
               if the program (eg su) is looking for it. */
            optdat->place = EMSG;
            if (strchr(ostr, '-') == NULL)
                return(-1);
            optdat->optopt = '-';
        }
    } else
        optdat->optopt = *optdat->place++;

    /* See if option letter is one the caller wanted... */
    if (optdat->optopt == ':' || (oli = strchr(ostr, optdat->optopt)) == NULL) {
        if (*optdat->place == 0)
            ++optdat->optind;
        if (optdat->opterr && *ostr != ':')
            (void)fprintf(stderr,
                          "%s: illegal option -- %c\n", _getprogname(nargv[0]),
                          optdat->optopt);
        return(BADCH);
    }

    /* Does this option need an argument? */
    if (oli[1] != ':') {
        /* don't need argument */
        optdat->optarg = NULL;
        if (*optdat->place == 0)
            ++optdat->optind;
    } else {
        /* Option-argument is either the rest of this argument or the
           entire next argument. */
        if (*optdat->place)
            optdat->optarg = CONST_CAST(char *, optdat->place);
        else if (nargc > ++optdat->optind)
            optdat->optarg = nargv[optdat->optind];
        else {
            /* option-argument absent */
            optdat->place = EMSG;
            if (*ostr == ':')
                return(BADARG);
            if (optdat->opterr)
                (void)fprintf(stderr,
                              "%s: option requires an argument -- %c\n",
                              _getprogname(nargv[0]), optdat->optopt);
            return(BADCH);
        }
        optdat->place = EMSG;
        ++optdat->optind;
    }
    return(optdat->optopt);            /* return option letter */
}

int
getopt_long(getopt_data_t* optdat,
            int argc, char *const argv[],
            const char *optstring,
            const struct option * longopts, int *longindex)
{
    char       *oli;            /* option letter list index */

    if (optdat->optreset || !*optdat->place) {     /* update scanning pointer */
        optdat->optreset = 0;

        if (optdat->optind >= argc) {
            optdat->place = EMSG;
            return -1;
        }

        optdat->place = argv[optdat->optind];

        if (optdat->place[0] != '-') {
            optdat->place = EMSG;
            return -1;
        }

        optdat->place++;

        if (optdat->place[0] && optdat->place[0] == '-' &&
            optdat->place[1] == '\0') {                       /* found "--" */
            ++optdat->optind;
            optdat->place = EMSG;
            return -1;
        }

        if (optdat->place[0] && optdat->place[0] == '-' && optdat->place[1]) {
            /* long option */
            size_t      namelen;
            int         i;

            optdat->place++;

            namelen = strcspn(optdat->place, "=");
            for (i = 0; longopts[i].name != NULL; i++) {
                if (strlen(longopts[i].name) == namelen
                    && strncmp(optdat->place, longopts[i].name, namelen) == 0) {
                    if (longopts[i].has_arg) {
                        if (optdat->place[namelen] == '=')
                            optdat->optarg = optdat->place + namelen + 1;
                        else if (optdat->optind < argc - 1) {
                            optdat->optind++;
                            optdat->optarg = argv[optdat->optind];
                        } else {
                            if (optstring[0] == ':')
                                return BADARG;
                            if (optdat->opterr)
                                fprintf(stderr,
                                        "%s: option requires an argument -- %s\n",
                                        argv[0], optdat->place);
                            optdat->place = EMSG;
                            optdat->optind++;
                            return BADCH;
                        }
                    } else {
                        optdat->optarg = NULL;
                        if (optdat->place[namelen] != 0) {
                            /* XXX error? */
                        }
                    }

                    optdat->optind++;

                    if (longindex)
                        *longindex = i;

                    optdat->place = EMSG;

                    if (longopts[i].flag == NULL)
                        return longopts[i].val;
                    else {
                        *longopts[i].flag = longopts[i].val;
                        return 0;
                    }
                }
            }

            if (optdat->opterr && optstring[0] != ':')
                fprintf(stderr,
                        "%s: illegal option -- %s\n", argv[0], optdat->place);
            optdat->place = EMSG;
            optdat->optind++;
            return BADCH;
        }
    }

    /* short option */
    optdat->optopt = (int) *optdat->place++;

    oli = strchr(optstring, optdat->optopt);
    if (!oli) {
        if (!*optdat->place)
            ++optdat->optind;
        if (optdat->opterr && *optstring != ':')
            fprintf(stderr,
                    "%s: illegal option -- %c\n", argv[0], optdat->optopt);
        return BADCH;
    }

    if (oli[1] != ':') {                           /* don't need argument */
        optdat->optarg = NULL;
        if (!*optdat->place)
            ++optdat->optind;
    } else {                           /* need an argument */
        if (*optdat->place)             /* no white space */
            optdat->optarg = optdat->place;
        else if (argc <= ++optdat->optind) {                       /* no arg */
            optdat->place = EMSG;
            if (*optstring == ':')
                return BADARG;
            if (optdat->opterr)
                fprintf(stderr,
                        "%s: option requires an argument -- %c\n",
                        argv[0], optdat->optopt);
            return BADCH;
        } else
            /* white space */
            optdat->optarg = argv[optdat->optind];
        optdat->place = EMSG;
        ++optdat->optind;
    }
    return optdat->optopt;
}
/*
 * vim:ts=4:sw=4:tw=78:et
 */
